/*
 * Copyright (c) 2019, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * WSO2 Inc. licenses this file to you under the Apache License,
 * Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package io.siddhi.core.query.table.store;

import io.siddhi.core.SiddhiAppRuntime;
import io.siddhi.core.SiddhiManager;
import io.siddhi.core.event.Event;
import io.siddhi.core.stream.input.InputHandler;
import io.siddhi.core.util.EventPrinter;
import org.apache.log4j.Logger;
import org.testng.AssertJUnit;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import java.sql.SQLException;

public class InsertIntoTestStoreTestCase {
    private static final Logger log = Logger.getLogger(InsertIntoTestStoreTestCase.class);

    @BeforeClass
    public static void startTest() {
        log.info("== Test Store INSERT tests started ==");
    }

    @AfterClass
    public static void shutdown() {
        log.info("== Test Store INSERT tests completed ==");
    }

    @Test
    public void insertIntoTestStoreTest1() throws InterruptedException, SQLException {
        //Configure siddhi to insert events data to table only from specific fields of the stream.
        log.info("insertIntoTestStoreTest1");
        SiddhiManager siddhiManager = new SiddhiManager();
        String streams = "" +
                "define stream StockStream (symbol string, price float, volume long); " +
                "@Store(type=\"testStoreContainingInMemoryTable\")\n" +
                "define table StockTable (symbol string, volume long); ";

        String query1 = "" +
                "@info(name = 'query1') " +
                "from StockStream\n" +
                "select symbol, volume\n" +
                "insert into StockTable ;";
        SiddhiAppRuntime siddhiAppRuntime = siddhiManager.createSiddhiAppRuntime(streams + query1);
        InputHandler stockStream = siddhiAppRuntime.getInputHandler("StockStream");
        siddhiAppRuntime.start();

        stockStream.send(new Object[]{"WSO2", 55.6f, 100L});
        stockStream.send(new Object[]{"IBM", 75.6f, 100L});
        Thread.sleep(1000);

        Event[] events = siddhiAppRuntime.query("" +
                "from StockTable ");
        EventPrinter.print(events);
        AssertJUnit.assertEquals(2, events.length);

        siddhiAppRuntime.shutdown();
    }

    @Test
    public void insertIntoTestStoreTest2() throws InterruptedException, SQLException {
        //Testing table creation with a compound primary key (normal insertion)
        log.info("insertIntoTestStoreTest2");
        SiddhiManager siddhiManager = new SiddhiManager();
        String streams = "" +
                "define stream StockStream (symbol string, price float, volume long); " +
                "@Store(type=\"testStoreContainingInMemoryTable\")\n" +
                "@PrimaryKey(\"symbol\", \"price\")" +
                "define table StockTable (symbol string, price float, volume long); ";

        String query = "" +
                "@info(name = 'query1') " +
                "from StockStream   " +
                "insert into StockTable ;";

        SiddhiAppRuntime siddhiAppRuntime = siddhiManager.createSiddhiAppRuntime(streams + query);
        InputHandler stockStream = siddhiAppRuntime.getInputHandler("StockStream");
        siddhiAppRuntime.start();

        stockStream.send(new Object[]{"WSO2", 55.6F, 100L});
        stockStream.send(new Object[]{"IBM", 75.6F, 100L});
        stockStream.send(new Object[]{"MSFT", 57.6F, 100L});
        stockStream.send(new Object[]{"WSO2", 58.6F, 100L});
        Thread.sleep(1000);

        Event[] events = siddhiAppRuntime.query("" +
                "from StockTable ");
        EventPrinter.print(events);
        AssertJUnit.assertEquals(4, events.length);

        siddhiAppRuntime.shutdown();
    }

    @Test
    public void insertIntoTestStoreTest3() throws InterruptedException, SQLException {
        //Testing table creation with a compound primary key (normal insertion)
        log.info("insertIntoTestStoreTest3");
        SiddhiManager siddhiManager = new SiddhiManager();
        String streams = "" +
                "define stream StockStream (symbol string, price float, volume long);\n";

        String table = "" +
                "@Store(type=\"testStoreContainingInMemoryTable\")\n" + // , @Cache(size="10")
                "@PrimaryKey(\"symbol\", \"price\")" +
                "define table StockTable (symbol string, price float, volume long);\n" +
                "@Store(type=\"testStoreContainingInMemoryTable\")\n" +
                "@PrimaryKey(\"symbol\", \"price\")" +
                "define table MyTable (symbol string, price float);\n";

        String query = "" +
                "@info(name = 'query1') " +
                "from StockStream   " +
                "select symbol, price " +
                "insert into MyTable ;";

        SiddhiAppRuntime siddhiAppRuntime = siddhiManager.createSiddhiAppRuntime(streams + table + query);
        InputHandler stockStream = siddhiAppRuntime.getInputHandler("StockStream");
        siddhiAppRuntime.start();

        stockStream.send(new Object[]{"WSO2", 55.6F});
        stockStream.send(new Object[]{"IBM", 75.6F});
        stockStream.send(new Object[]{"MSFT", 57.6F});
        stockStream.send(new Object[]{"WSO2", 58.6F});
        Thread.sleep(1000);

        Event[] events = siddhiAppRuntime.query("" +
                "from MyTable ");
        EventPrinter.print(events);
        AssertJUnit.assertEquals(4, events.length);

        siddhiAppRuntime.shutdown();
    }
}
